package ap.proof;

import ap.terfor.{Conjunction, Formula, TermOrder}
import ap.util.{Debug, Logic}

object AndTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(subtrees : Seq[ProofTree], order : TermOrder) : ProofTree =
    new AndTree(subtrees, order)
  
}

class AndTree private (val subtrees : Seq[ProofTree],
                       protected val order : TermOrder) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(AndTree.AC, !subtrees.isEmpty)
  //////////////////////////////////////////////////////////////////////////////
  
  lazy val closingConstraint : Formula =
    ConstraintSimplifier(
      Conjunction.conj(for (t <- subtrees.elements) yield t.closingConstraint,
                       order),
                order)

  lazy val stepPossible : Boolean =
    Logic.exists(for (subtree <- subtrees.elements) yield subtree.stepPossible)

  def update(newSubtrees : Seq[ProofTree]) : ProofTree = AndTree(newSubtrees, order)
    
  //////////////////////////////////////////////////////////////////////////////

  private def indent(str : String, prefix : String) =
    prefix + str.replaceAll("\n", "\n" + prefix)
    
  override def toString : String = {
    "AndTree (" + closingConstraint + ") {\n" +
      (for (t <- subtrees) yield indent("" + t, "  ")).reduceLeft(
         (s1 : String, s2 : String) => s1 + "\n/\\\n" + s2) +
    "\n}"
  }

}
