package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.substitutions.{Substitution, ComposeSubsts}
import ap.util.Debug

object SubstTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(subtree : ProofTree, subst : Substitution,
            order : TermOrder, subtreeOrder : TermOrder)
                                              : ProofTree = subtree match {
    case subtree : SubstTree =>
      SubstTree(subtree.subtree,
                ComposeSubsts(subst sortBy subtree.subtreeOrder,
                              subtree.subst,
                              subtree.subtreeOrder),
                order,
                subtree.subtreeOrder)
    case _ => new SubstTree(subtree, subst, order, subtreeOrder)
  }
  
}

/**
 * <code>ProofTreeOneChild</code> that applies a substitution to the closing
 * constraint of its <code>subtree</code>
 */
class SubstTree private (val subtree : ProofTree,
                         val subst : Substitution,
                         protected val order : TermOrder,
                         protected val subtreeOrder : TermOrder)
      extends ProofTreeOneChild {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(SubstTree.AC,
                   (subst isSortedBy subtreeOrder) &&
                   (subtree match {
                     // nested substitutions are supposed to be contracted
                     case subtree : SubstTree => false
                     case _ => true
                    }))
  //////////////////////////////////////////////////////////////////////////////

  lazy val closingConstraint : Formula =
    (order sort subst(subtree.closingConstraint))

  def update(newSubtree : ProofTree) : ProofTree =
    SubstTree(newSubtree, subst, order, subtreeOrder)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    "SubstTree (" + subst + ") (" + closingConstraint + ") {\n" + subtree + "\n}"
  }
}
