package ap.terfor.substitutions;


import ap.util.{Debug, Logic}

object VariableSubst {
  
  private val AC = Debug.AC_SUBSTITUTIONS
  
}

/**
 * Substitute the variable starting from index <code>offset</code> with the
 * terms in <code>replacements</code>. I.e., <code>VariableTerm(offset)</code>
 * is going to be replaced with <code>replacements(0)</code>, etc. All other
 * variables above <code>offset+replacements.size</code> are shifted downwards
 * by <code>replacements.size</code>
 */
class VariableSubst(offset : Int,
                    replacements : Seq[Term],
                    protected [substitutions] val order : TermOrder)
      extends SimpleSubstitution {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(VariableSubst.AC,
                   (offset >= 0) &&
                   Logic.forall(for (t <- replacements.elements)
                                yield (order isSortingOf t)))
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution = {
    // TODO: could be optimised by caching
    val shiftSubst = new VariableShiftSubst(0, num, order)
    val newReplacements = for (t <- replacements) yield shiftSubst(t)
    new VariableSubst(offset + num, newReplacements, order)
  }
      
  protected[substitutions] def applyToVariable(v : VariableTerm) : Term = v match {
    case VariableTerm(i) =>
      if (i < offset) {
      	v
      } else if (i - offset < replacements.size) {
        replacements(i - offset)
      } else {
        VariableTerm(i - replacements.size)
      }
  }

  protected[substitutions] def applyToConstant(c : ConstantTerm) : Term = c

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    Logic.forall(for (VariableTerm(i) <- t.variables.elements) yield (i < offset))

  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : VariableSubst =
    new VariableSubst(offset,
                      for (t <- replacements) yield (newOrder sort t),
                      newOrder)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    val strings = for ((t, i) <- replacements.elements.zipWithIndex)
                  yield ("" + VariableTerm(i+offset) + "|->" + t)
    val substs = if (strings.hasNext)
                   strings.reduceLeft((s1 : String, s2 : String) => s1 + " | " + s2)
                 else
                   ""
    "[" + substs + "]"
  }
}
